/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/

#include "config.h"
#include "GeirsPanic.h"

#include <string>
#include <cstring>
#include <vector>
#include <fstream>
#include <limits>   /* quiet_NaN */
#include <libgen.h>

using namespace std ;
using namespace CCfits ;

/* Use an algorithm which blanks (for the case requested) within the loop
* that copies data from the subwindows to the super-frame. If undefined, the 
* initial subwindows are blanked first (looking only near the gaps of the fused array)
* and then copied - including the BLANKS - to the super-frame.
* Does make a difference of roughly 10 percent in speed and simplified handling of DATAMIN
*/
// #define SLOW_TRIM

/** 
* @param[in] fitsname The name of the FITS file to be read.
* @param[in] fitstpl The name of the FITS file to be created.
*  This may be the empty string ("") indicating that the file name is to be determined with
*  a standard rule.
* @param[in] gap The gap between the chips of the mosaic in units of pixels.
* @param[in] chpsz The edge length of the single chip in pixel units.
*   This would be 2048 for Hawaii2 and Hawaii2-RG, or 1024 for Hawaii1.
* @since 2012-10-23
* @author Richard J. Mathar
*/
// GeirsPanic::GeirsPanic(char *fitsIname, char *fitsOname) : iname(fitsIname), oname(fitsOname)
GeirsPanic::GeirsPanic(char *fitsIname, const char *fitsOname, int gap, int chpsz) 
 : iname(fitsIname) ,ws(fitsIname,gap,chpsz), oname(fitsOname)
{
	// ifits = new FITS(iname, CCfits::Read) ;
	/* create the window information, read all their pixels
	ws(fitsIname,gap,csiz) ;
	*/
	ofits = 0 ;
#ifdef DEBUG
	FITS::setVerboseMode(true) ;
#endif
} /* GeirsPanic::GeirsPanic */

#if 0
/** Destructor.
*/
GeirsPanic::~GeirsPanic()
{
	/* remove the allocated resoures with the ctor of ifits within the ctor
	*/
	delete ifits ;
} /* GeirsPanic::dtor */
#endif

/** Modify the FITS header and optionally rotate/flip the image.
* @param[in] doTrim If true, remove (blanken) the 4 pixels at the edges of each detector.
* @param[in] doMimax If true, write DATAMIN, DATAMAX, DATAMED etc to the header.
* @param[in] gap The size of the crossing gap (pixels) in the mosaic.
* @param[in] chpsz The size of a single chip (pixels) of the mosaic.
* @param[in] north The direction angle of North in the current image (+90=up, +180 = left)
* @param[in] doflip If true, flip the image. 
*   The main idea is to trigger a swap of the Eastern side relative to North, if 
*   previously found to be wrong.
*   There is of course an effect of this option on the direction of North in the
*   final image, unless this was exactly +90 or -90.
* @param[in] doaltaz If true, derive ALT, AZ and PARANG from latitude, HA and DEC.
*   Warning: this will only produce correct values if the keywords HA, DEC and GEOPOS_L
*   are degrees in the original header, or HA and DEC are given in hex-decimal colon-separated notation.
* @param[in] doGeo If true, add OBSGEO-[XYZ] to the keywords.
* @param[in] verbose If true, write more verbose output.
* @param[in] remov If true, the initial file (or the initial file set if windowed) is removed on exit.
*    This is always done if requested, and may lead to loss of data if anything is wrong with the
*    new file that is created.
* @since 2012-10-18
* @todo Divert windows into header units instead inserting the gaps into the single image.
* @author Richard J. Mathar
*/
void GeirsPanic::exec(bool doTrim, bool doMimax, double north , bool doflip, bool doaltaz, bool doGeo,
	bool verbose, bool remov)
{
	if ( ws.ws.size() == 0 )
	{
		cerr << "No input files detected. No output files produced." << endl ;
		return ;
	}

	/* normalize to positive north angular value
	*/
	if ( north < 0.)
		north += 360.0 ;

	/* quartrot is the number of 90 degrees clockwise rotations we wish to rotate the image
	* If >45 and < 135, quartrot=0. So North up is roughly the intended standard orientation
	*/
	int quartrot =3 ;
	if ( north >= 45.0 && north < 135.0)
		quartrot= 0 ;
	if ( north >= 135.0 && north < 225.0)
		quartrot= 1 ;
	else if ( north >= 225.0 && north < 315.0)
		quartrot= 2 ;
	north -= quartrot*90.0 ;

	if (verbose && quartrot != 0)
		cout << "Image will be rotated " << quartrot << " times 90 degrees clockwise " << endl ;

	/* define the location and orientation of the final window. This updates all
	* the trimsize[] entries in the subwindows and also the global trimsize in the set.
	*/
	ws.win2win(quartrot,doflip) ;

	int refIdx =0 ;
	/* check that this first member acutally exists: already done above
	*/
	// Window & refWin = ws.ws[refIdx] ;
	Window & refWin = *ws.ws[refIdx] ;
	long bits = refWin.bits ;

	/* two or three components for images or cubes, respectively
	*/
	vector<long> Vdatasec ;
	for(int i=0 ; i < refWin.naxes ; i++)
		Vdatasec.push_back(1) ;


	PHDU & iex = refWin.ifits->pHDU() ;
	iex.readAllKeys() ;

	/* use zeros for the blank value on output. Any other value is  biased
	* since the integer images are usually subtracted and may contain negative values.
	*/
	int blankI = -999999 ;
	unsigned short int blankUS = 0 ;
	short int blankS = 0 ;
	float blankF = numeric_limits<float>::quiet_NaN() ;

	/* readAllKeys does not read the BLANK keyword because this is not in the s_iKeywordCategories of
	* HDU.cxx. But we still can read it (if present), which will throw a NoSuchKeyword() exception
	* if the BLANK is actually missing. We simply try to keep the original value
	* if present.
	*/
	try
	{
		int nulvalue ;
		iex.readKey(string("BLANK"),nulvalue) ;
		blankI = blankUS = blankS = nulvalue ;
	}
	catch(HDU::NoSuchKeyword ex)
	{
	}

	/* one plane if flat image, 1 or larger (=third dimension) if a data cube
	*/
	const int planes = (refWin.naxes == 3) ? iex.axis(2) : 1 ;

	long datasec[3] ;
	datasec[0] = ws.trimsize[1]-ws.trimsize[0] ;
	datasec[1] = ws.trimsize[3]-ws.trimsize[2] ;
	datasec[2] = planes ;
	if (verbose)
		cout << "Image with gap is " << datasec[0] << " by " << datasec[1] << " by " << datasec[2] << " pixels." << endl ;

	int chpsz = refWin.chpsz ;

	/* create a default output name if the current one is empty
	*/
	if ( oname.length() == 0 )
	{
cout << "...input " << iname << endl ;
		char * inamptr = new char[iname.length()+1] ;
		strcpy(inamptr,iname.c_str()) ;
cout << "...input " << inamptr << endl ;
		oname = string(dirname(inamptr))+ "/" + ws.ws[0]->canonName() ;
		delete [] inamptr ;
	}

	if (verbose)
		cout << "...Output to " << oname << endl ;

	/* create a single image for single image input, and cube for cube input
	*/
	ofits = new FITS(oname, bits,refWin.naxes,datasec) ;
	PHDU & oex = ofits->pHDU() ;

	map<String, Keyword*> & allkw = iex.keyWord() ;

	/* gather right ascens, declination and hour angle in case we want to calculate the alt-az related coordinates
	*/
	double ra=0. ;
	double dec=0. ;
	double ha=0. ;
	/* longitude, latitude and altitude in case we want to use the OBSGEO-XYZ */
	double topocen[3] ;

	/* copy most header keyword lines. This does not add BZERO and BSCAL to the list
	* of keywords (with a philosophy that these have already been applied on reading).
	* It also does not show BLANK, if set in the original file (which is apparently also eaten by
	* the underlying cfitsio reader).
	* Do not copy the pseudo-void keywords left by the standard GEIRS output.
	*/
	map<String, Keyword*>::iterator kitr = allkw.begin() ;
	for ( ; kitr != allkw.end() ; kitr++)
	{
		const string & knam = (*kitr).first;
		// cout << "...input " << knam << endl ;
		if ( knam != "CTYPE1" && knam != "CTYPE2" && knam != "CRPIX1" && knam != "CRPIX2"
			&& knam != "CDELT1" && knam != "CDELT2" && knam != "CRVAL1" && knam != "CRVAL2"
			&& knam != "SAVEAREA" && knam != "TRIMSIZE" && knam != "DATASEC" )
		{
			const Keyword * thisk = (*kitr).second ;
			oex.addKey(thisk) ;
			string tmp ;
			/* for RA, DEC and HA try to autodetect hex-strings versus pure numbers.
			*/
			if ( knam == "RA" )
			{
				if ( thisk->keytype() == Tstring )
				{
					thisk->value(tmp) ;
					ra = hex2deg(tmp,false) ;
				}
				else
					thisk->value(ra) ;
			}
			else if ( knam == "DEC" )
			{
				if ( thisk->keytype() == Tstring )
				{
					thisk->value(tmp) ;
					dec = hex2deg(tmp,true) ;
				}
				else
					thisk->value(dec) ;
			}
			else if ( knam == "HA" )
			{
				if ( thisk->keytype() == Tstring )
				{
					thisk->value(tmp) ;
					ha = hex2deg(tmp,false) ;
				}
				else
					thisk->value(ha) ;
			}

			else if ( doGeo || doaltaz)
			{
				if ( knam == "TELLONG" || knam == "OBSGEO-L" )
					thisk->value(topocen[0]) ;
				else if ( knam == "TELLAT" || knam == "OBSGEO-B" )
					thisk->value(topocen[1]) ;
				else if ( knam == "TELALT" || knam == "OBSGEO-H" )
					thisk->value(topocen[2]) ;
			}
		}
	}

	/* Output arrays. only one of these will actually be used, matching the
	* integer/short/float type of the first window.
	*/
	valarray<float> Vfarr  ;
	valarray<int> Viarr ;
	valarray<unsigned short> Vusarr ;
	valarray<short> Vsarr ;

	/* datamin and datamax over the subwindows, potentially excluding the reference pixels
	* at the edges of the chips. Note that the computation is not postponed
	* to a Vusarr.min(), Vusarr.max() etc later on because these functions would
	* also cover the BLANK value in the output arrays.
	*/
	double mima[2] ;
	mima[0] = numeric_limits<double>::max() ;
	mima[1] = -numeric_limits<double>::max() ;

	/* allocate the output array for that particular type of data
	*/
	switch(bits)
	{
	case USHORT_IMG:
#if 0
		/* work around a bug in older g++ compilers like 4.3.
		*/
		Vusarr = valarray<unsigned short>(blankUS,datasec[0]*datasec[1]*datasec[2]) ;
#else
		Vusarr.resize(datasec[0]*datasec[1]*datasec[2],blankUS) ;
#endif
		break;
	case SHORT_IMG:
#if 0
		/* work around a bug in older g++ compilers like 4.3.
		*/
		Vsarr = valarray<short>(blankUS,datasec[0]*datasec[1]*datasec[2]) ;
#else
		Vsarr.resize(datasec[0]*datasec[1]*datasec[2],blankUS) ;
#endif
		break;
	case LONG_IMG:
#if 0
		/* work around a bug in older g++ compilers like 4.3.
		*/
		Viarr = valarray<int>(blankI,datasec[0]*datasec[1]*datasec[2]) ;
#else
		Viarr.resize(datasec[0]*datasec[1]*datasec[2],blankI) ;
#endif
		break;
	case FLOAT_IMG:
#if 0
		/* work around a bug in older g++ compilers like 4.3.
		*/
		Vfarr = valarray<float>(blankF,datasec[0]*datasec[1]*datasec[2]) ;
#else
		Vfarr.resize(datasec[0]*datasec[1]*datasec[2],blankF) ;
#endif
		break;
	case BYTE_IMG:
	case LONGLONG_IMG:
	case DOUBLE_IMG:
		break;
	}

	int coo[2] ;
	/* outer loop over all planes (or only 1 plane if a flat image)
	*/
	for(int pl=0 ; pl < planes ; pl++)
	{
		switch(bits)
		{
		case USHORT_IMG:
			for(int i=0 ; i < ws.ws.size() ; i++)
			{
				int origSiz[2] ;
				Window & subw = *ws.ws[i] ;
				origSiz[0] = subw.sizeX() ;
				origSiz[1] = subw.sizeY() ;
#if not defined SLOW_TRIM
				if ( doTrim)
					subw.blankRef(blankUS) ;
#endif
				for(int y=0 ; y < origSiz[1] ; y++)
				{
					/* targ[0] and targ[1] are two consecutive coordinates in the sequential
					* array of values in the output super-window. Because the original walking
					* through the subwindow is also in a consecutive fashion, the estimate targ[2]
					* for the next placement is just the linear extrapolation of targ[0] and targ[1]
					* for any combination of rotations and flips.
					* Except the cases where the walk through the subwindow meets a line break or
					* meets the edge of the chip and moves on within another chip of a mosaic.
					* If this occurs (tested by remainder check in the super-frame coordinates),
					* we interrupt that linear extrapolation scheme and start another one.
					*/
					int targ[3] = {0,0,0} ;
					for(int x=0 ; x < origSiz[0] ; x++)
					{
						if ( (x+subw.trimsizeFus[0]) % subw.chpsz == 0 )
							targ[0] = targ[1]=0 ;
	
						if ( targ[0] == 0 )
						{
							/* targ[0]=0 indicate that we have not yet seen two consecutive
							* previous locations and cannot take targ[2] as the next position
							*/
							coo[0] = x ;
							coo[1] = y ;
							/* rotate in the global coordinate system of the spliced windows
							*/
							subw.cooGeirsWinToSplicRot(coo, quartrot, doflip) ;
							/* translate relativ to the rotated union of the superframe
							*/
							coo[0] -= ws.trimsize[0] ;
							coo[1] -= ws.trimsize[2] ;
							targ[2] = coo[0]+datasec[0]*coo[1] +pl*datasec[0]*datasec[1] ;
						}
	
	
#if defined SLOW_TRIM
						if ( ! doTrim || ! subw.isRef(x+subw.trimsizeFus[0],y+subw.trimsizeFus[2]) )
						{
#endif
							const double tmp = (Vusarr[targ[2]] = subw.usarr[x+origSiz[0]*y+origSiz[0]*origSiz[1]*pl] );
							mima[0] = min(mima[0],tmp) ;
							mima[1] = max(mima[1],tmp) ;
#if defined SLOW_TRIM
						}
#endif
						targ[0] = targ[1] ;
						targ[1] = targ[2] ;
						targ[2] = 2*targ[1]-targ[0] ;
					}
				}
			}
			break;
		case SHORT_IMG:
			for(int i=0 ; i < ws.ws.size() ; i++)
			{
				int origSiz[2] ;
				Window & subw = *ws.ws[i] ;
				origSiz[0] = subw.sizeX() ;
				origSiz[1] = subw.sizeY() ;
#if not defined SLOW_TRIM
				if ( doTrim)
					subw.blankRef(blankS) ;
#endif
				for(int y=0 ; y < origSiz[1] ; y++)
				{
					/* targ[0] and targ[1] are two consecutive coordinates in the sequential
					* array of values in the output super-window. Because the original walking
					* through the subwindow is also in a consecutive fashion, the estimate targ[2]
					* for the next placement is just the linear extrapolation of targ[0] and targ[1]
					* for any combination of rotations and flips.
					* Except the cases where the walk through the subwindow meets a line break or
					* meets the edge of the chip and moves on within another chip of a mosaic.
					* If this occurs (tested by remainder check in the super-frame coordinates),
					* we interrupt that linear extrapolation scheme and start another one.
					*/
					int targ[3] = {0,0,0} ;
					for(int x=0 ; x < origSiz[0] ; x++)
					{
						if ( (x+subw.trimsizeFus[0]) % subw.chpsz == 0 )
							targ[0] = targ[1]=0 ;
	
						if ( targ[0] == 0 )
						{
							/* targ[0]=0 indicate that we have not yet seen two consecutive
							* previous locations and cannot take targ[2] as the next position
							*/
							coo[0] = x ;
							coo[1] = y ;
							/* rotate in the global coordinate system of the spliced windows
							*/
							subw.cooGeirsWinToSplicRot(coo, quartrot, doflip) ;
							/* translate relativ to the rotated union of the superframe
							*/
							coo[0] -= ws.trimsize[0] ;
							coo[1] -= ws.trimsize[2] ;
							targ[2] = coo[0]+datasec[0]*coo[1] +pl*datasec[0]*datasec[1] ;
						}
	
	
#if defined SLOW_TRIM
						if ( ! doTrim || ! subw.isRef(x+subw.trimsizeFus[0],y+subw.trimsizeFus[2]) )
						{
#endif
							const double tmp = (Vsarr[targ[2]] = subw.sarr[x+origSiz[0]*y+origSiz[0]*origSiz[1]*pl] );
							mima[0] = min(mima[0],tmp) ;
							mima[1] = max(mima[1],tmp) ;
#if defined SLOW_TRIM
						}
#endif
						targ[0] = targ[1] ;
						targ[1] = targ[2] ;
						targ[2] = 2*targ[1]-targ[0] ;
					}
				}
			}
			break;
		case LONG_IMG:
			for(int i=0 ; i < ws.ws.size() ; i++)
			{
				int origSiz[2] ;
				Window & subw = *ws.ws[i] ;
				origSiz[0] = subw.sizeX() ;
				origSiz[1] = subw.sizeY() ;
#if not defined SLOW_TRIM
				if ( doTrim)
					subw.blankRef(blankI) ;
#endif
				for(int y=0 ; y < origSiz[1] ; y++)
				{
					/* targ[0] and targ[1] are two consecutive coordinates in the sequential
					* array of values in the output super-window. Because the original walking
					* through the subwindow is also in a consecutive fashion, the estimate targ[2]
					* for the next placement is just the linear extrapolation of targ[0] and targ[1]
					* for any combination of rotations and flips.
					* Except the cases where the walk through the subwindow meets a line break or
					* meets the edge of the chip and moves on within another chip of a mosaic.
					* If this occurs (tested by remainder check in the super-frame coordinates),
					* we interrupt that linear extrapolation scheme and start another one.
					*/
					int targ[3] = {0,0,0} ;
					for(int x=0 ; x < origSiz[0] ; x++)
					{
						if ( (x+subw.trimsizeFus[0]) % subw.chpsz == 0 )
							targ[0] = targ[1]=0 ;
	
						if ( targ[0] == 0 )
						{
							/* targ[0]=0 indicate that we have not yet seen two consecutive
							* previous locations and cannot take targ[2] as the next position
							*/
							coo[0] = x ;
							coo[1] = y ;
							/* rotate in the global coordinate system of the spliced windows
							*/
							subw.cooGeirsWinToSplicRot(coo, quartrot, doflip) ;
							/* translate relativ to the rotated union of the superframe
							*/
							coo[0] -= ws.trimsize[0] ;
							coo[1] -= ws.trimsize[2] ;
							targ[2] = coo[0]+datasec[0]*coo[1]  +pl*datasec[0]*datasec[1];
						}
	
	
#if defined SLOW_TRIM
						if ( ! doTrim || ! subw.isRef(x+subw.trimsizeFus[0],y+subw.trimsizeFus[2]) )
						{
#endif
							const double tmp = (Viarr[targ[2]] = subw.iarr[x+origSiz[0]*y+origSiz[0]*origSiz[1]*pl] );
							mima[0] = min(mima[0],tmp) ;
							mima[1] = max(mima[1],tmp) ;
#if defined SLOW_TRIM
						}
#endif
						targ[0] = targ[1] ;
						targ[1] = targ[2] ;
						targ[2] = 2*targ[1]-targ[0] ;
					}
				}
			}
			break;
		case FLOAT_IMG:
			for(int i=0 ; i < ws.ws.size() ; i++)
			{
				int origSiz[2] ;
				Window & subw = *ws.ws[i] ;
				origSiz[0] = subw.sizeX() ;
				origSiz[1] = subw.sizeY() ;
#if not defined SLOW_TRIM
				if ( doTrim)
					subw.blankRef(blankI) ;
#endif
				for(int y=0 ; y < origSiz[1] ; y++)
				{
					/* targ[0] and targ[1] are two consecutive coordinates in the sequential
					* array of values in the output super-window. Because the original walking
					* through the subwindow is also in a consecutive fashion, the estimate targ[2]
					* for the next placement is just the linear extrapolation of targ[0] and targ[1]
					* for any combination of rotations and flips.
					* Except the cases where the walk through the subwindow meets a line break or
					* meets the edge of the chip and moves on within another chip of a mosaic.
					* If this occurs (tested by remainder check in the super-frame coordinates),
					* we interrupt that linear extrapolation scheme and start another one.
					*/
					int targ[3] = {0,0,0} ;
					for(int x=0 ; x < origSiz[0] ; x++)
					{
						if ( (x+subw.trimsizeFus[0]) % subw.chpsz == 0 )
							targ[0] = targ[1]=0 ;
	
						if ( targ[0] == 0 )
						{
							/* targ[0]=0 indicate that we have not yet seen two consecutive
							* previous locations and cannot take targ[2] as the next position
							*/
							coo[0] = x ;
							coo[1] = y ;
							/* rotate in the global coordinate system of the spliced windows
							*/
							subw.cooGeirsWinToSplicRot(coo, quartrot, doflip) ;
							/* translate relativ to the rotated union of the superframe
							*/
							coo[0] -= ws.trimsize[0] ;
							coo[1] -= ws.trimsize[2] ;
							targ[2] = coo[0]+datasec[0]*coo[1] +pl*datasec[0]*datasec[1] ;
						}
	
	
#if defined SLOW_TRIM
						if ( ! doTrim || ! subw.isRef(x+subw.trimsizeFus[0],y+subw.trimsizeFus[2]) )
						{
#endif
							const double tmp = (Vfarr[targ[2]] = subw.farr[x+origSiz[0]*y+origSiz[0]*origSiz[1]*pl] );
							mima[0] = min(mima[0],tmp) ;
							mima[1] = max(mima[1],tmp) ;
#if defined SLOW_TRIM
						}
#endif
						targ[0] = targ[1] ;
						targ[1] = targ[2] ;
						targ[2] = 2*targ[1]-targ[0] ;
					}
				}
			}
			break;
		case BYTE_IMG:
		case LONGLONG_IMG:
		case DOUBLE_IMG:
			break;
		}  /* end of switch bitpix */
	} /* end of loop over the planes (more than 1 if cubes) */

	if ( doMimax)
	{
		oex.addKey("DATAMIN",mima[0]," [adu] minimum value in image") ;
		oex.addKey("DATAMAX",mima[1]," [adu] maximum value in image") ;
	}

	oex.addKey("CUNIT1","deg","WCS units along axis 1") ;
	oex.addKey("CUNIT2","deg","WCS units along axis 2") ;
	oex.addKey("CTYPE1","RA---TAN","WCS axis 1") ;
	oex.addKey("CTYPE2","DEC--TAN","WCS axis 2") ;
	oex.addKey("CRVAL1",ra,"[deg] RA in center") ;
	oex.addKey("CRVAL2",dec,"[deg] DEC in center") ;

	/* original location of reference pixel was the center of the mosaic.
	* In full-frame images, this falls onto a non-existing integer pixel of the fused image,
	* so to get it to the center of the gap, use a transformation to the left and right and
	* take the mean....  detsize[]/2-1 will be a pixel just left to the middle, and detsize[]/2 just to the right.
	*/
	int cooCtr[2][2] ;
	cooCtr[0][0] = refWin.detsize[1]/2 -1 -refWin.trimsizeFus[0];
	cooCtr[0][1] = refWin.detsize[3]/2 -1 -refWin.trimsizeFus[2];
	cooCtr[1][0] = cooCtr[0][0]+1 ;
	cooCtr[1][1] = cooCtr[0][1]+1 ;
	for(int d=0 ; d < 2 ; d++)
	{
		refWin.cooGeirsWinToSplicRot(cooCtr[d], quartrot, doflip) ;
		cooCtr[d][0] -= ws.trimsize[0] ;
		cooCtr[d][1] -= ws.trimsize[2] ;
	}

	oex.addKey("CRPIX1",(cooCtr[0][0]+cooCtr[1][0])/2," RA and DEC center along axis 1") ;
	oex.addKey("CRPIX2",(cooCtr[0][1]+cooCtr[1][1])/2," RA and DEC center along axis 2") ;

	if ( doflip)
		north = 180-north ;
	/* rotation on the detector, from deg to radians
	*/
	north *= M_PI/180.0 ;
	float pixsc = refWin.pixsc ;
	oex.addKey("CD1_1",-pixsc*sin(north)," [deg/px] WCS matrix diagonal") ;
	oex.addKey("CD2_2",pixsc*sin(north)," [deg/px] WCS matrix diagonal") ;
	oex.addKey("CD1_2",pixsc*cos(north)," [deg/px] WCS matrix outer diagonal") ;
	oex.addKey("CD2_1",pixsc*cos(north)," [deg/px] WCS matrix outer diagonal") ;

	if ( doGeo)
	{
		double geoc[3] ;
		wgs84(topocen,geoc) ;
		oex.addKey("OBSGEO-X",geoc[0]," [m] WGS84 geocentric x coordinate ") ;
		oex.addKey("OBSGEO-Y",geoc[1]," [m] WGS84 geocentric y coordinate ") ;
		oex.addKey("OBSGEO-Z",geoc[2]," [m] WGS84 geocentric z coordinate ") ;
	}

	if ( doaltaz)
	{
		double aap[3] ;
		altAzPar(ha,dec,topocen[1],aap) ;
		oex.addKey("ALT",aap[0]," [deg] altitude above topocentr. horizon") ;
		oex.addKey("AZ",aap[1]," [deg] azimuth, S =0, W=90") ;
		oex.addKey("PARANG",aap[2]," [deg] parallactic angle") ;
	}

	switch(bits)
	{
	case USHORT_IMG:
		oex.addKey("BLANK",blankUS,"") ;
		ofits->flush() ;
		oex.write(Vdatasec,datasec[0]*datasec[1]*datasec[2],Vusarr, & blankUS) ;
		break;
	case SHORT_IMG:
		oex.addKey("BLANK",blankS,"") ;
		ofits->flush() ;
		oex.write(Vdatasec,datasec[0]*datasec[1]*datasec[2],Vsarr, & blankS) ;
		break;
	case LONG_IMG:
		oex.addKey("BLANK",blankI,"") ;
		ofits->flush() ;
		oex.write(Vdatasec,datasec[0]*datasec[1]*datasec[2],Viarr, & blankI) ;
		break;
	case FLOAT_IMG:
		/* in floating point images, the NaN represents the invalid BLANK
		* and is not written according tothe FITS recommendations.
		*/
		ofits->flush() ;
		oex.write(Vdatasec,datasec[0]*datasec[1]*datasec[2],Vfarr, & blankF) ;
		break;
	case BYTE_IMG:
	case LONGLONG_IMG:
	case DOUBLE_IMG:
		break;
	}

	/** If any keywords were changed or added, write a HISTORY line
	* keeping track of the template file, and remove the CHECKSUM keyword
	* if present.
	*/

	/* Keep the date of the original creation intact...
	* iex.writeDate() ;
	*/
	time_t now = time(NULL) ;
	string nowLoc = ctime(&now) ;
	if ( getenv("LOGNAME") )
		nowLoc += string(getenv("LOGNAME")) ;
	if ( getenv("DOMAIN") )
		nowLoc += string(getenv("DOMAIN")) ;
	iex.writeHistory("Rewritten by "__FILE__" from "+ refWin.ifits->name() + " " +nowLoc);

	/* remove the CHECKSUM and DATASUM if present (because these become invalid)
	*/
	try
	{
		iex.deleteKey("CHECKSUM") ;
		iex.deleteKey("DATASUM") ;
	}
	catch(HDU::NoSuchKeyword ex2)
	{
		/* may throw an exception if the keyword is already
		* absent, in which we're not interested
		*/
	}

	/* dtor time is when the file is actually saved to the disk...
	*/
	delete ofits ;

	/* finally (so this is safest if any exception was thrown above,
	* delete the input files if requested
	*/
	if ( remov)
		ws.rmFiles(verbose) ;
} /* GeirsPanic::exec */

/** Convert geographic longitude, latitude and sea leavel to geocentric Cartesian coords.
* @param[in] lolatalt The longitude (lambda, degrees), latitude (phi, degrees) and altitude (meters) of the site.
   The latitude is the topocentric one, not the geocentric one.
* @param[out] xyz The x, y and z coordinates in the Greenwhich centered coordinates (meters).
* @since 2012-10-25
*/
void GeirsPanic::wgs84(const double lolatalt[3], double xyz[3])
{
	/* radius, semi-major axis (m)
	*/
	static const double rhoe = 6378137.0 ;
	/* static const oneoverf = 298.257223563 ; inverse flattening, not needed */
	/* first eccentricity squared (!)
	*/
	static const double esqur = 6.69437999014e-3 ;
	/* cosine and sine phi and lambda */
	double csphilam[2][2] ;
	csphilam[0][0] = cos(lolatalt[1]*M_PI/180.0) ; /* cosine phi */
	csphilam[0][1] = sin(lolatalt[1]*M_PI/180.0) ; /* sine phi */
	csphilam[1][0] = cos(lolatalt[0]*M_PI/180.0) ; /* cosine lambda */
	csphilam[1][1] = sin(lolatalt[0]*M_PI/180.0) ; /* sine lambda */
	const double N = rhoe/sqrt(1.-esqur*csphilam[0][1]*csphilam[0][1]) ;
	xyz[0] = (N+lolatalt[2])*csphilam[0][0]*csphilam[1][0] ;
	xyz[1] = (N+lolatalt[2])*csphilam[0][0]*csphilam[1][1] ;
	xyz[2] = (N*(1.-esqur)+lolatalt[2])*csphilam[0][1] ;
} /* wgs84 */

/** Derive altitude (complementary zenith), azimuth and paralactic angle.
* Note that this will fail (drastically) if any of the units in the original header
* for the hour angle (HA), declination (DEC) or latitude (OBSGEO-L) were not degrees.
* @param[in] ha Hour angle at the epoch (deg)
* @param[in] dec Declination at the epoch (deg).
* @param[in] phi  Observer's latitude (deg).
* @param[out] aap The altitude (deg), azimuth (deg, South to West) and parallactic angle (deg)
* @since 2012-10-25
*/
void GeirsPanic::altAzPar(double ha, double dec, double phi, double aap[3])
{
	/* convert to radians */
	ha *= M_PI/180.0 ;
	dec *= M_PI/180.0 ;
	phi *= M_PI/180.0 ;
	const double sina = sin(dec)*sin(phi)+cos(dec)*cos(ha)*cos(phi) ;
	aap[0] = asin(sina) ;
	const double cosasinA = cos(dec)*sin(ha) ;
	const double cosacosA = -sin(dec)*cos(phi)+cos(dec)*cos(ha)*sin(phi) ;
	aap[1] = atan2(cosasinA,cosacosA) ;

	/* product cos(delta)*sin(parallactic)
	*/
	const double cosdsinp = cos(phi)*sin(aap[1]) ;

	/* product cos(delta)*cos(parallactic)
	*/
	const double cosdcosp = cos(phi)*cos(aap[1]) *sin(aap[0])+sin(phi)*cos(aap[0]);
	aap[2] = atan2(cosdsinp,cosdcosp) ;

	/* convert values from radian to degree on return
	*/
	aap[0] *= 180.0/M_PI ;
	aap[1] *= 180.0/M_PI ;
	aap[2] *= 180.0/M_PI ;
} /* altAzpar */

/** Convert a sedecimal-String to degrees.
* @param[in] hexstr A string with optional sign, integer number, colon, integer number, colon
*   and integer or floating point number.
* @param[in] isdeg If true, assume that the string is in DD:MM:SS.ss format.
*   If false the string is in HH:MM:SS.ss format.
*   This implies an additional factor 15 to move on to degrees.
*   If the string contains a sign (plus or minus), this parameter is ignored
*   and the function assumes degrees.
* @return The value in degrees.
* @since 2012-11-13
*/
double GeirsPanic::hex2deg(const string & hexstr, bool isdeg)
{
	/* find first colon, last colon, and optional sign */
	size_t fcol = hexstr.find_first_of(":") ;
	size_t lcol = hexstr.find_last_of(":") ;
	size_t sig = hexstr.find_first_of("+-") ;

	/* If a sign was found before the first colon, assume that the argument
	* was given in degrees, not hours.
	*/
	if ( sig != string::npos && sig < fcol)
		isdeg = true ;

	double deg ;
	// if ( fcol != string::npos && lcol != string::npos && dot != string::npos )
	if ( fcol != string::npos && lcol != string::npos )
	{
		int dm[2] ;

		/* first (optionally signed) number between start and fcol
		*/
		sscanf(hexstr.substr(0,fcol).c_str(),"%d",&dm[0]) ;

		/* second number between fcol and lcol
		*/
		sscanf(hexstr.substr(fcol+1,lcol-fcol-1).c_str(),"%d",&dm[1]) ;

		double s ;
		/* third number after lcol
		*/
		sscanf(hexstr.substr(lcol+1).c_str(),"%lf",&s) ;

		/* unsigned absolute value
		*/
		deg  = fabs(dm[0]) + dm[1]/60.0 + s/3600.0 ;
		if ( ! isdeg)
			deg *= 15.0 ;
		if ( dm[0] < 0)
			/* add sign back to result */
			deg *= -1 ;
	}
	else
	{
		/* if no or only 1 colon is found, assume this is a simple number, already in degrees
		*/
		sscanf(hexstr.c_str(),"%lf",&deg) ;
	}
	return deg ;
} /* hex2deg */
