/*
* MPIA - GEIRS project
*
* $Header$
*
*/


#include "config.h"

// #include <string>
// #include <cstring>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>

#include "Histo.h"

using namespace std ;

/** Constructor.
* Empty histogram. The main purpose of this construction
* is to allow allocations of vectors of histograms.
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histo::Histo() : cts(), cumCts()
{
} /* Histo::Histo */

/** Constructor.
* @param arr The array of values to be binned.
* @param Nbin The number of bins into which the values are split.
* @param range The minimum and maximum value in the first and last bin.
* @todo Check that Nbin is larger than zero etc.
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histo::Histo(const valarray<float> & arr, const int Nbin, const double * range, string & nam) : cts(), cumCts(), name(nam)
{
	init(arr, Nbin, range) ;
} /* Histo::Histo */

/** Constructor.
* The range of the bins is automatically set to reach out to the smallest
* and largest value in the array.
* @param arr The array of values to be binned.
* @param Nbin The number of bins into which the values are split.
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histo::Histo(const valarray<float> & arr, const int Nbin, string &nam) : cts(), cumCts(), name(nam)
{
	double range[2] ;
	range[0] = arr.min() ;
	range[1] = arr.max() ;
	/* decrease range by half a bin size such that the lowest value
	* is in the middle of the lowest bin. Equivalent for the largest bin
	*/
	const double diff = (range[1]-range[0])/Nbin ;
	range[0] -= diff/2. ;
	range[1] += diff/2. ;
	init(arr,Nbin,range) ;
} /* Histo::Histo */

/** Constructor.
* The number of bins is set equal to the square root of the number of values in the array.
* The range of the bins is automatically set to reach out to the smallest
* and largest value in the array.
* @param arr The array of values to be binned.
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histo::Histo(const valarray<float> & arr, string & nam) : cts(), cumCts(), name(nam)
{
	int Nbin = (int)(sqrt(arr.size())+0.5 );
	double range[2] ;
	range[0] = arr.min() ;
	range[1] = arr.max() ;
	/* decrease range by half a bin size such that the lowest value
	* is in the middle of the lowest bin. Equivalent for the largest bin
	*/
	const double diff = (range[1]-range[0])/Nbin ;
	range[0] -= diff/2. ;
	range[1] += diff/2. ;
	init(arr,Nbin,range) ;
} /* Histo::Histo */

/* Destructor. Superfluous so far.
Histo::~Histo()
{
}
*/

/** Principal part of the constructor.
* The statistics (minimum and maximum and the percentiles) is computed
* over the entire range of the arr-values that are not NaN's, not only
* over the values within the interval specified by the range argument.
* @param arr The array of values to be binned.
* @param Nbin The number of bins into which the values are split.
* @param range The minimum and maximum value in the first and last bin.
* @todo Check that Nbin is larger than zero etc.
* @since 2013-01-28
* @since 2013-02-22 percentiles added
* @since 2013-06-07 NaN's ignored.
* @since 2013-12-04 cumulative bins added
* @author Richard J. Mathar
*/
void Histo::init( const valarray<float> & arr, const int Nbin, const double range[])
{
	/* initialize all bins to zero (in case there has an older
	* instance not been reallocating these...
	*/
	for(int b=0 ; b < Nbin ; b++)
	{
		cts.push_back(0) ;
		cumCts.push_back(0) ;
	}
	/* define leftmost bin and bin widths
	*/
	strt = range[0] ;
	stride = (range[1]-range[0])/Nbin ;

	/* for each value in the array which is in the range,
	* determine it's bin location by distance towards leftmost bin modulo the stride.
	* Do not count NaN's.
	*/
	int N = arr.size() ;
	for(int v=0 ; v < N ; v++)
	{
		if ( isnan(arr[v]) )
			;
		else if ( arr[v] >= range[0] && arr[v] < range[1]) 
		{
			cts[(int)((arr[v]-strt)/stride) ] ++ ;
		}
	}
	for(int b=0 ; b < Nbin ; b++)
	{
		if ( b == 0 )
			cumCts[b] = cts[b] ;
		else
			cumCts[b] = cumCts[b-1]+cts[b] ;
	}
	

	/* count the values which are not NaN and which are not outside
	* our binning range. So nonblanks will almost always be less than arr.size().
	*/
	int nonblanks = 0 ;
	for(int b=0 ; b < cts.size() ; b++)
		nonblanks += cts[b] ;

	/* compute percentiles corresponding to mean and +-1 to 3 sigma
	* erf(n/sqrt(2)) = 0.00269979606326 (and 1 minus this), 0.0455002638964 (and 1 minus this)
	* and 0.317310507863 (and 1 minus this)
	* Percentile (from 0 to 1) = (index+1/2)/N, where N is the number of data points.
	*/
	static double erfc[] = {0.00269979606326, 0.0455002638964, 0.317310507863, 0.5,
		0.682689492137, 0.954499736104, 0.997300203937} ;

	/* sort a local copy of the data, not the original data
	* At the same time, compile the minimum and maximum of the valid
	* (non-NaN) entries.
	*/
	float *arrS = new float[N] ;

	/* index of the destination varray (counts how many of the
	* original data, within or not within the bins, were non-NaN's
	*/
	int vdest =0 ;

	for(int vsrc=0 ; vsrc< N; vsrc++)
	{
		if ( isnan(arr[vsrc]) ) 
			;
		else
		{
			if ( vdest == 0 )
				mimax[0] = mimax[1] = arr[vsrc] ;
			else
			{
				mimax[0] = min(mimax[0], (double) arr[vsrc]) ;
				mimax[1] = max(mimax[1], (double) arr[vsrc]) ;
			}
			/* accumulate the sum over all values in mimax[2],
			* and divide through the count later to get the mean.
			*/
			mimax[2] += arr[vsrc] ;
			arrS[vdest++] = arr[vsrc] ;
		}
	}

	/* mimax[2] is the mean of valid values (including those outside the bins
	*/
	if ( vdest)
		mimax[2] /= vdest ;

	qsort(arrS, vdest, sizeof(float), fsort) ;

	for(int p =0 ; p < 7 ; p++)
	{
		int idx = (int)(erfc[p]*vdest-0.5) ;

		/* might do a linear interpolation, but this is overkill in this application
		*/
		if ( idx >= 0)
			perc[p] = (arrS[idx]+arrS[idx+1])/2. ;
		else
			perc[p] = arrS[0] ;
	}

	delete [] arrS ;
} /* init */

/** Determine smallest and largest value in all elements of the array.
* NaN-values are not taken into account (ignored). In that sense, this function
* is an improvement relative to the undefined result if min() or max() of the 
* array are taken as defined in the STL.
* @param[in] arr The array to be scanned.
* @param[out] mimax The minimum in [0], the maximum in [1].
*  If the array does not contain values that are not NaN's, both values are unchanged.
* @return The number of non-NaN's in the array.
* @since 2013-06-07
*/
int Histo::valRange( const valarray<float> & arr, double mimax[2])
{
	int nscanned =0 ;
        for ( int i=0 ; i < arr.size() ; i++)
        {
		if ( isnan(arr[i]) )
			;
		else if ( nscanned == 0 )
		{
			mimax[0] = mimax[1] = arr[i] ;
			nscanned++ ;
		}
		else
		{
			mimax[0] = min(mimax[0], (double) arr[i]) ;
			mimax[1] = max(mimax[1], (double) arr[i]) ;
			nscanned++ ;
		}
        }
	return nscanned ;
} /* valRange */


/*!*******************************
* @brief Support of the qsort call in init().
* @param v1 The first constant to be compared.
* @param v2 The second constant to be compared.
* @return -1 if v1 is less than v2, +1 if v1 is larger than v2, 0 if both are equal.
*/
int fsort(const void * v1, const void * v2)
{
	if ( *(float*)v1 < *(float*)v2)
		return -1;
	else if ( *(float*)v1 > *(float*)v2)
		return 1;
	else
		return 0 ;
}

/*!****************************************
* @brief Create the bins in xy gnuplot format.
* For each number in the histogram, a short horizontal line is
* produced, such that concatenation of these horizontal lines with
* straight line segments produces a step function along the x axis.
* @param[in] fname The file to be generated.
*/
void Histo::dumpFil(const char *fname)
{
	ofstream of(fname) ;
	for(int b= 0 ; b < cts.size() ; b++)
	{
		of << (strt + b*stride) << " " << cts[b] << endl ;
		of << (strt + (b+1)*stride) << " " << cts[b] << endl ;
	}
} /* dumpFil */
