/*
* MPIA - GEIRS project
*
* $Header$
*
*/

#include <iostream>
#include <fstream>
#include <cmath>
#include <cstring>

#include "config.h"
#include "Histos.h"

using namespace std ;

/** Constructor.
* Empty histogram. The main purpose of this construction
* is to allow allocations of vectors of 
* histograms.
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histos::Histos() : hs(), contName() 
{
} /* Histos::Histos */

/** Constructor.
* @param arr The array of values to be binned.
* @param iname The associated list of names, one per arr component.
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histos::Histos( vector< valarray<float> > & arr, vector<string> & iname) : hs(), contName()
{
	/* determine min-max range
	*/
	double range[2];
	valRange(arr,range) ;


	int Nbin = countRange(arr) ;
	Nbin = (int) sqrt(Nbin) ;

	init(arr,Nbin,range,iname) ;
} /* Histos::Histos */

/** Constructor.
* @param arr The array of values to be binned.
* @param Nbin A number of bins to be used.
*   If less than 1, the program depicts a default on its own.
* @param iname Name of the files associated with the arr.
* @param conts A global simple name for the entire collection.
*   Mainly for use with display headers and the like.
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histos::Histos( vector< valarray<float> > & arr, int Nbin, vector<string> & iname, string & conts) 
	: hs(), contName(conts)
{
	/* determine min-max range
	*/
	double range[2];
	valRange(arr,range) ;

	if ( Nbin <= 0 )
	{
		Nbin = countRange(arr) ;
		Nbin = (int) sqrt(Nbin) ;
	}

	init(arr,Nbin,range,iname) ;
} /* Histos::Histos */

/** Constructor.
* @param arr The array of values to be binned.
* @param Nbin A number of bins to be used.
*   If less than 1, the program depicts a default on its own.
* @param iname Name of the files associated with the arr.
* @param conts A global simple name for the entire collection.
*   Mainly for use with display headers and the like.
* @param[in] range
* @since 2013-01-28
* @author Richard J. Mathar
*/
Histos::Histos( vector< valarray<float> > & arr, int Nbin, vector<string> & iname, string & conts, double range[2]) 
	: hs(), contName(conts)
{
	if ( isnan(range[0]) || isnan(range[1]) ) 
	{
		double truRange[2] ;
		valRange(arr,truRange) ;
		if ( isnan(range[0]) )
			range[0] = truRange[0] ;
		if ( isnan(range[1]) )
			range[1] = truRange[1] ;
	}

	if ( Nbin <= 0 )
	{
		Nbin = countRange(arr) ;
		Nbin = (int) sqrt(Nbin) ;
	}

	init(arr,Nbin,range,iname) ;
} /* Histos::Histos */

/** Determine smallest and largest value in all elements of the arrays.
* NaN-values are not taken into account (ignored).
* @param[in] arr The arrays to be scanned.
* @param[out] mimax The minimum in [0], the maximum in [1]
*  If the array arr has zero size, the mimax[] are unchanged on return.
*/
void Histos::valRange( vector< valarray<float> > & arr, double mimax[2])
{
	/* need to check size to avoid core dumps while accesssing arr[0]
	*/
	if ( arr.size() > 0 )
	{
		/* locmimax stores the local extrema of arr[0], arr[1],...
		*/
		double locmimax[2] ;
		/* n is the total number of valid (non-NaN's) in the
		* arr[0], arr[1]... seen so far
		*/
		int n = Histo::valRange(arr[0],locmimax) ;
		if ( n > 0 )
			/* actually got some non-NaN's in the  first element
			* so use that result and copy it into mimax
			*/
			memcpy(mimax, locmimax, 2*sizeof(double)) ;

		for(int h= 1 ; h < arr.size() ; h++)
		{
			int nloc = Histo::valRange(arr[h],locmimax) ;
			if ( nloc > 0 )
			{
				/* actually got some non-NaN's in the element
				* so use that result and copy it into mimax
				*/
				if ( n == 0 )
				{
					memcpy(mimax, locmimax, 2*sizeof(double)) ;
				}
				else
				{
					mimax[0] = min(mimax[0],locmimax[0]) ;
					mimax[1] = max(mimax[1],locmimax[1]) ;
				}
			}
			n += nloc ;
		}
	}
} /* valRange */

/** 
* @param[in] arr The arrays to be scanned.
* @return
*/
int Histos::countRange( vector< valarray<float> > & arr)
{
	int autoNbin = 0 ;
	/* determine min-max range
	*/
	if ( arr.size() > 0 )
	{
		autoNbin = max(autoNbin, (int)(arr[0].size())) ;
		for(int h= 1 ; h < arr.size() ; h++)
			autoNbin = max(autoNbin, (int)(arr[h].size())) ;
	}
	return autoNbin ;
} /* Histos::countRange */

/* Destructor. Superfluous so far.
Histos::~Histos()
{
}
*/

/** Principal part of the constructor
* @param arr The array of values to be binned.
* @param Nbin The number of bins into which the values are split.
* @param range The minimum and maximum value in the first and last bin.
* @todo Check that Nbin is larger than zero etc.
* @since 2013-01-28
* @author Richard J. Mathar
*/
void Histos::init( vector< valarray<float> > & arr, const int Nbin, const double * range, vector<string> & iname)
{
	for(int b=0 ; b < arr.size() ; b++)
	{
		hs.push_back( Histo(arr[b],Nbin, range, iname[b]) ) ;
		/* report median and 1 sigma percentiles
		*/
		cout << iname[b] << ":" << endl ;
		cout << "median: " << hs[b].perc[3]  ;
		cout << " +" << hs[b].perc[4]-hs[b].perc[3] << " -" << hs[b].perc[3]-hs[b].perc[2] << " (1 sigma)" ;
		cout << " +" << hs[b].perc[5]-hs[b].perc[3] << " -" << hs[b].perc[3]-hs[b].perc[1] << " (2 sigma)" ;
		cout << " +" << hs[b].perc[6]-hs[b].perc[3] << " -" << hs[b].perc[3]-hs[b].perc[0] << " (3 sigma)" << endl ;
	}
} /* init */

/** Create ASCII file in xy gnuplot format.
* @param[in] fname The file to be generated.
* @since 2013-01-29
* @since 2013-12-04 cumulative sum column added.
* @author Richard J. Mathar
*/
void Histos::dumpFil(const char *fname)
{
	// std::cout << __FILE__ << " " << __LINE__ << " " << fname << " " << hs.size() << endl ;
	ofstream of(fname) ;
	if ( hs.size() > 0)
	{
		double str = hs[0].strt ;
		double step = hs[0].stride ;
		/* loop over the horizontal axis of the histogram
		*/
		for(int b= 0 ; b < hs[0].cts.size() ; b++)
		{
			/* internal loop over all histograms, side by
			* side in the line
			*/
			of << (str + b*step)  ;
			for(int h= 0 ; h < hs.size() ; h++)
				of << " " << hs[h].cts[b]  ;
			for(int h= 0 ; h < hs.size() ; h++)
				of << " " << hs[h].cumCts[b]  ;
			of << endl ; 
			of << (str + (b+1)*step) ;
			for(int h= 0 ; h < hs.size() ; h++)
				of << " " << hs[h].cts[b]  ;
			for(int h= 0 ; h < hs.size() ; h++)
				of << " " << hs[h].cumCts[b]  ;
			of << endl ; 
		}
	}
} /* dumpFil */

/** Generate a gnuplot shell command list.
*@param[in] fdname The file that contains the data to be plotted.
*@param[in] fpltname The file with the gnuplot commands.
*@param[in] doLogHist Print a logarithmic, not a linear vertical scale
*@param[in] epsout Print diagram into the EPS file, not on screen.
* @since 2013-01-29
* @author Richard J. Mathar
*/
void Histos::gnuplot(const char *fdname, const char *fpltname, const bool doLogHist, string epsout)
{
	// std::cout << __FILE__ << " " << __LINE__ << " " << fdname << " " << fpltname << endl ;
	ofstream of(fpltname) ;
	if ( hs.size() > 0)
	{
		of << "set title \"" << contName << "\"" << endl ;
#if 0
		of << "set multiplot" << endl ;
		of << "set ylabel \"counts \"" << endl ;
		of << "set style data lines" << endl ;

		of << "set size 1.,0.5" << endl ;
		of << "set origin 0.,0.5" << endl ;
		of << "plot " ;
		for(int h= 0 ; h < hs.size() ; h++)
		{
			of << "\"" << fdname << "\" using 1:" << (h+2) << " title \"" << hs[h].name << "\"" ;
			if ( h < hs.size()-1 )
				of << ","  ;
		}
		of << endl ;

		of << "set xlabel \"ADU\"" << endl ;
		of << "set logscale y" << endl ;
		of << "set origin 0.,0." << endl ;

		of << "plot " ;
		for(int h= 0 ; h < hs.size() ; h++)
		{
			of << "\"" << fdname << "\" using 1:" << (h+2) << " title \"" << hs[h].name << "\"" ;
			if ( h < hs.size()-1 )
				of << ","  ;
		}
		of << endl ;

		of << "unset multiplot" << endl ;
		of << "pause -1" << endl ;
#else
		of << "set ylabel \"counts \"" << endl ;
		of << "set style data lines" << endl ;

		if ( doLogHist)
			of << "set logscale y" << endl ;

		if ( epsout.length() > 0)
		{
			std::cout << "printing into " << epsout << endl ;
			of << "set size 1.,0.7" << endl ;
			of << "set terminal postscript color portrait" << endl ;
			of << "set output \"" << epsout << "\"" << endl ;
		}

		of << "set xlabel \"ADU\"" << endl ;
		of << "plot " ;
		for(int h= 0 ; h < hs.size() ; h++)
		{
			of << "\"" << fdname << "\" using 1:" << (h+2) << " title \"" << hs[h].name << "\"" ;
			if ( h < hs.size()-1 )
				of << ","  ;
		}
		of << endl ;
		if ( epsout.length() == 0)
			of << "pause -1" << endl ;
#endif

	}
} /* gnuplot */
