#include "config.h"

#include <algorithm>
#include <limits>

#include <CCfits/CCfits>

#include "PixImage.h"
#include "Bbox2D.h"

using namespace CCfits ;

/*!*************************************
* @brief Create an empty image with all values set to zero.
* @param[in] ncols The number of columns in the image.
*  The same as the number of pixels along x.
* @param[out] nrows The number of rows in the image.
*  The same as the number of pixels along y.
*/
PixImage::PixImage(const int ncols, const int nrows) : rows(nrows), cols(ncols)
{
	//vals = valarray<double>(0.,rows*cols) ;
	vals = valarray<double>(rows*cols) ;
	// work around a bug in gcc 4.3
	//vals.resize(rows*cols) ;
} /* ctor */

/*!*************************************
* @brief Create an image of zero size (without pixels).
*/
PixImage::PixImage() : rows(0), cols(0)
{
} /* default ctor */

/*!*************************************
* @brief Read the pixel values from a slice of a FITS cube.
* @param fit The FITS object with at least one image plane in the Primary HDU.
* @param slic The 0-based index into the FITS cube.
*/
PixImage::PixImage(FITS & fit, int slic)
{
	rows = cols = 0;

	PHDU & p = fit.pHDU() ;
	switch( p.axes() )
	{
	case 2 :
		slic=0 ;
		cols = p.axis(0) ;
		rows = p.axis(1) ;
		break;
	case 3 :
		if ( slic < 0 )
			slic=0 ;
		if ( slic >= p.axis(2) )
			slic= p.axis(2)-1 ;
		cols = p.axis(0) ;
		rows = p.axis(1) ;
		break;
	default:
		cerr << "No image in primary HDU of " << fit.name() << endl ;
		slic = -1 ;
	}
	if ( slic >= 0)
	{
		p.read(vals, 1L+slic*rows*cols,rows*cols) ;
		// cout << __FILE__ << " " << __LINE__ << " siz " << fit.name() << " " << vals.size() << endl ;
	}
} /* ctor */

/*!*************************************
* Destructor. Release the allocated memory areas.
*/
PixImage::~PixImage()
{
} /* dtor */

/*!****************************************
* Generate a hull (an encompassing integer rectangle) of the image of a pixel
* @param[in] rt The rotation-translation to be applied to the pixel.
* @param[in] row The row index of the pixel.
* @param[in] col The column index of the pixel.
* @param[in] bbrot The boundary box of the rotated entire original array.
* @param[out] idxReg The index region of the image of the pixel under the operation.
*  idxReg[0][] refers to the lower left corner, idxReg[1][] to the upper right.
*  idxReg[][0] refers to the x coordinates, idxReg[][1] to the y coordinates.
*  idx[1][] is non-incluse (as usual in C/C++/Java): The pixel (row,col) will
*  cover parts of idxReg[0][0]..idxReg[0][1]-1 along x and idxReg[1][0]..idxRe[1][1]-1
*  along y after rotation.
* @return The rotated pixel in the original coordinate system.
*/
Square2D PixImage::hull(const RotTrans2D & rt, const int row, const int col, int idxReg[2][2], const Bbox2D & bbrot) const
{
	/* Given its x-range col..col+1 and its y-range row..row+1,
	* define the (floating point) bounding box of the image of this pixel.
	Bbox2D bbpix( Point2D(col,row), Point2D(col+1.,row+1)) ;
	Bbox2D pixrot = bbpix.apply(rt) ;
	*/

	Square2D sqp( Point2D(col,row), Point2D(col+1.,row)) ;
	Square2D sqprot = sqp.apply(rt) ;
	Bbox2D pixrot = sqprot ;

	/* re-define bounding box in the coordinate syste of bbrot,
	* where bbrot.limits[0] defines the new integer coordinate origin.
	pixrot.limits[0] = bbrot.limits[0].to( pixrot.limits[0]) ;
	pixrot.limits[1] = bbrot.limits[0].to( pixrot.limits[1]) ;
	*/
	pixrot.limits[0] -= bbrot.limits[0] ;
	pixrot.limits[1] -= bbrot.limits[0] ;

	/* build floor of pixrot.limits[0] and ceiling of pixrot.limits[1]
	*/
	idxReg[0][0] = max(0,(int)pixrot.limits[0].coord[0]) ; // lower left x, rounding down
	idxReg[0][1] = max(0,(int)pixrot.limits[0].coord[1]) ; // lower left y, rounding down
	idxReg[1][0] = (int)(2.+pixrot.limits[1].coord[0]) ; // upper right x, rounding up
	idxReg[1][1] = (int)(2.+pixrot.limits[1].coord[1]) ; // upper right y, rounding up

	return sqprot ;
} /* hull */

/*!*******************************************
* @brief Read/get the value of the intensity at a specific pixel
* @param[in] row
* @param[in] col
* @return The current contents at that pixel
*/
double PixImage::at(const int row, const int col) const
{
	/* linear flat array with index of the columns fastest
	*/
	return vals[idx(row,col)] ;
}

/*!*******************************************
* @brief Obtain the flat index of a (x,y) pair of indices.
* @param[in] row
* @param[in] col
* @return The 0-based index col+(number of columns)*row.
* @since 2013-07-06
*/
int PixImage::idx(const int row, const int col) const
{
	return col + cols*row ;
}


/*!*******************************************
* @brief Generate the bounding box of this pixel array.
* @return The quadrangular bounding box.
*/
PixImage::operator Bbox2D() const
{
	return Bbox2D( Point2D(0.,0.), Point2D(cols,rows)) ;
} /* operator Bbox2D() */

/*!*******************************************
* @brief Generate a FITS file with this image as the primary HDU.
* @param[in] foutname
* @return The FITS file with a dummy header.
* @since 2013-07-06
*/
FITS * PixImage::toFits(const char *foutname) const
{
	long axs[3] ;
	axs[0] = cols ;
	axs[1] = rows ;
	axs[2] = 1 ;
	FITS * fout = new FITS(String(foutname),FLOAT_IMG,2,axs) ;
	PHDU & p = fout->pHDU() ;
	p.write(1L, cols*rows, vals) ;
	return fout ;
} /* operator FITS() */

/*!****************************************
* Generate a quadrangle of pixels by rotation and translation of a indexed pixel.
* @todo Carry over NaN pixel values (BLANK) to the result.
* @since 2013-07-06
*/
#if 0
PixImage PixImage::apply(const RotTrans2D & rt) const
#else
PixImage * PixImage::apply(const RotTrans2D & rt) const
#endif
{
	/* figure out the coordinates of the old corner points
	* (0,0) up to (cols,rows) in integer units under the tranformation.
	*/
	Bbox2D bb = *this ;

	/* Rotate the original bounding box and find/define the
	* bounding box of the rotated image.
	*/
	Bbox2D bbrot = bb.apply(rt) ;

	/* Since binning is not included (yet) in this library, the horizontal and
	* vertical number of pixels is the ceiled values of the bbox coordinates.
	*/
	int imgCols = bbrot.limits[1].coord[0]-bbrot.limits[0].coord[0]+1. ;
	int imgRows = bbrot.limits[1].coord[1]-bbrot.limits[0].coord[1]+1. ;

	/* this implies that the pixel column number 0 in the new image starts
	* at the x-value bbrot.limits[0].coord[0] and ends at +1 higher up.
	*/

	/* Define the bare bone geometry the rotated image with all pixel
	* values initially zero.
	*/
#if 0
	PixImage img(imgCols, imgRows) ;
#else
	PixImage * img = new PixImage(imgCols, imgRows) ;
#endif

	/* loop over all pixels of this and copy the associated (partial)
	* flux into the pixels of img.
	*/
	for(int row =0 ; row < rows; row++)
		for(int col =0 ; col < cols ; col++)
		{
			/* the rotated image of this pixel covers a cluster
			* of pixels in the coordinate system of the new image
			* with a small (not more than 3x3) boundary box in that new image.
			*/
			int hullIdx[2][2] ;
			Square2D rotpix = hull(rt, row,col,hullIdx,bbrot) ;

			/* relocate the 4 rotated vertices to the origin defined by bbrot
			*/
			for(int v=0 ;v < 4 ; v++)
				rotpix.vert[v] -= bbrot.limits[0] ;

			/* The original pixel contributes to all pixels with
			* a portion of light that is proportional to the portion of the
			* area of the overlap between the rotate pixel and the pixel in
			* the rotated image.
			*/
			for(int colI = hullIdx[0][0] ; colI < hullIdx[1][0] && colI < imgCols ; colI++)
				for(int rowI = hullIdx[0][1] ; rowI < hullIdx[1][1] && rowI < imgRows; rowI++)
				{
					/* unit square in the target image
					*/
					Square2D imgI( Point2D(colI,rowI) , Point2D(colI+1.,rowI),
							Point2D(colI+1,rowI+1), Point2D(colI,rowI+1.) );

					vector<Tria2D> ov = rotpix.intersect(imgI) ;
					/* The homogeneous illumination of this pixel is applied to
					* all broken pieces of the triangles. Sum the area of all
					* triangular subareas that contribute to pixel (colI,rowI)
					* and multiply by the value of pixel (row,col) in the
					* original image.
					*/
					double ar=0. ;
					for( int t=0 ; t < ov.size() ; t++)
					{
						ar += ov[t].area() ;
					}
#if 0
					img.vals[img.idx(rowI,colI)] += at(row,col)*ar ;
#else
					img->vals[img->idx(rowI,colI)] += at(row,col)*ar ;
#endif
				}
		}

#if 0
	/* debugging: a test output of a quarter of the final image
	*/
	for(int row=0 ; row < imgRows/2 ; row++)
	{
		for(int col=0 ; col < imgCols/2 ; col++)
			cout << col << " " << row << " " << img.at(row,col) << endl ;
		cout << endl ;
	}
#endif

	return img ;
} /* apply */
