
#include "config.h"

#ifdef HAVE_CMATH
#include <cmath>
#else
#ifdef HAVE_MATH_H
#include <math.h>
#endif
#endif


#include "RotTrans2D.h"


/*!*************************************
* Create a rotation-translation given the angle and translation vector.
* @param ang The angle (ccw, radians)
* @param x The x coordinate of the translation.
* @param y The y coordinate of the translation
*/
RotTrans2D::RotTrans2D(double ang, double x, double y) : angle(ang), trans(x,y)
{
#ifdef HAVE_SINCOS
        sincos(angle,&sc[0], &sc[1]) ;
#else   
        sc[0] = sin(ang) ;
        sc[1] = cos(ang) ;
#endif  
} /* ctor */

/*!*************************************
* Create a rotation-translation given the angle and translation vector.
* @param ang The angle (ccw, radians)
* @param t The translation vector represented by a point in the original coordinates.
*/
RotTrans2D::RotTrans2D(double ang, const Point2D & t) : angle(ang), trans(t)
{
#ifdef HAVE_SINCOS
        sincos(angle,&sc[0], &sc[1]) ;
#else   
        sc[0] = sin(ang) ;
        sc[1] = cos(ang) ;
#endif  
} /* ctor */

/*!*************************************
* Create a rotation-translation given the angle and translation vector.
* The initial rotation is specified by an angle and a fixed point of the rotation.
* @param ang The angle (ccw, radians)
* @param fixpt The fixed point for the rotation.
* @param t The translation vector represented by a point in the original coordinates.
*/
RotTrans2D::RotTrans2D(double ang, const Point2D & fixpt, const Point2D & t)
	: angle(ang)
{
	/* The rotation around a generic point is implicitly a
	* translation of this point to the origin, a rotation around the
	* origin, and finally a translation by the negative amount:
	* ximg = x  -fixpt.x
	* yimg = y  -fixpt.y
	* ximg2 = cos(angle)*ximg -sin(angle)*yimg
	* yimg2 = sin(angle)*ximg +cos(angle)*yimg
	* ximg3 = ximg2 +fixpt.x
	* yimg3 = yimg2 +fixpt.y
	* Writing ximg3 and yimg3 as a function of x,y, fixpt and the angle gives
	* ximg3 = [cos(angle)*x-sin(angle)*y] +(1-cos(angle))*fixpt.x +sin(angle)*fixpt.y
	* yimg3 = [sin(angle)*x+cos(angle)*y] -sin(angle)*fixpt.x +(1-cos(angle))*fixpt.y
	// Maple
	// ximg := x-fptx ;
	// yimg := y-fpty ;
	// ximg2 := cos(angle)*ximg -sin(angle)*yimg ;
	// yimg2 := sin(angle)*ximg +cos(angle)*yimg ;
	// ximg3 := expand(ximg2+fptx+tx) ;
	// yimg3 := expand(yimg2+fpty+ty);
	*/
#ifdef HAVE_SINCOS
        sincos(angle,&sc[0], &sc[1]) ;
#else   
        sc[0] = sin(ang) ;
        sc[1] = cos(ang) ;
#endif  

	double fullt[2] ;
	fullt[0] = (1.-sc[1])*fixpt.coord[0] +sc[0]*fixpt.coord[1] +t.coord[0] ;
	fullt[1] = (1.-sc[1])*fixpt.coord[1] -sc[0]*fixpt.coord[0] + t.coord[1] ;
	trans = Point2D(fullt[0],fullt[1]) ;
} /* ctor */

/*!*************************************
* Create a dummy rotation-translation (which is not rotating or translating).
* This results in a rotation by an angle of zero and translation
* by a vector of zero length.
*/
RotTrans2D::RotTrans2D() : angle(0.0), trans()
{
        sc[0] = 0.0 ;
        sc[1] = 1.0 ;
} /* ctor */

/*!*************************************
* Apply the transformation to a point.
* @param pt The point to be rotated-translated.
* @return The rotated coordinates.
*/
Point2D RotTrans2D::apply(const Point2D &pt) const
{
	/* multply from the left by the matrix 
	* (cos(alpha), - sin(alpha))
	* (sin(alpha), + cos(alpha))
	* and then translate by trans.
	*/
	const double x = sc[1]* pt.coord[0] - sc[0]*pt.coord[1] + trans.coord[0] ;
	const double y = sc[0]* pt.coord[0] + sc[1]*pt.coord[1] + trans.coord[1] ;
	return Point2D(x,y) ;
} /* apply */

/*!****************************************
* @brief Generate the inverse operation of this.
* @return The left inverse operation M', defined by M'*this = identity,
*  or the right inverse operation M' defined by this*M'=identity. In the
* operator products, the latter one (the one to the right of the multiplication
* symbol) is executed first.
*/
RotTrans2D RotTrans2D::inv() const
{
	/* (i) for the left inverse:
	* Need to find M' that satisfies M'*M=I where M'=R'+t' is the
	* operation to be generated (returned) and M=R+t is this.
	* R' rotates by -angle, the opposite direction.
	* If M is characterised by the action
	* ximg = cos(angle)*x -sin(angle)*y + tx;
	* yimg = sin(angle)*x +cos(angle)*y + ty;
	* and if M' is characterised by the action with -angle and primed t,
	* ximg2 = cos(-angle)*ximg -sin(-angle)*yimg + tx' ;
	* yimg2 = sin(-angle)*ximg +cos(-angle)*yimg + ty';
	* then the operation M followed by the operation M' is characterized by
	* ximg2 = cos(angle)*tx+sin(angle)*ty+tx'+x;
	* yimg2 = -sin(angle)*tx+cos(angle)*ty+ty'+y.
	//    In Maple:
	//    alphinv := -alpha ;
	//    xprime := cos(alpha)*x-sin(alpha)*y+tx ;
	//    yprime := sin(alpha)*x+cos(alpha)*y+ty ;
	//    xtwise := cos(alphinv)*xprime-sin(alphinv)*yprime+txprime ;
	//    ytwise := sin(alphinv)*xprime+cos(alphinv)*yprime+typrime ;
	//    simplify(expand(xtwise),trig) ;
	//    simplify(expand(ytwise),trig) ;
	* To enforce ximg2=x and yimg2=y in these 2 equations, wee need to have
	* 0 = cos(angle)*tx+sin(angle)*ty+tx'
	* 0 = -sin(angle)*tx+cos(angle)*ty+ty'
	* tx' = -cos(angle)*tx-sin(angle)*ty
	* ty' = sin(angle)*tx-cos(angle)*ty
	*/

	/* (ii) for the right inverse:
	* Need to find M' that satisfies M*M'=I where M'=R'+t' is the
	* operation to be generated (returned) and M=R+t is this.
	* R' rotates by -angle, the opposite direction.
	* If M' is characterised by the action
	* ximg = cos(-angle)*x -sin(-angle)*y + tx';
	* yimg = sin(-angle)*x +cos(-angle)*y + ty';
	* and if M is characterised by the action
	* ximg2 = cos(angle)*ximg -sin(angle)*yimg + tx ;
	* yimg2 = sin(angle)*ximg +cos(angle)*yimg + ty;
	* then the operation M' followed by the operation M is characterized by
	* ximg2 = cos(angle)*tx'-sin(angle)*ty'+tx+x;
	* yimg2 = sin(angle)*tx'+cos(angle)*ty'+ty+y.
	//    In Maple:
	//    alphinv := -alpha ;
	//    xprime := cos(-alpha)*x-sin(-alpha)*y+txprime;
	//    yprime := sin(-alpha)*x+cos(-alpha)*y+typrime ;
	//    xtwise := cos(alpha)*xprime-sin(alpha)*yprime+tx ;
	//    ytwise := sin(alpha)*xprime+cos(alpha)*yprime+ty ;
	//    simplify(expand(xtwise),trig) ;
	//    simplify(expand(ytwise),trig) ;
	* To enforce ximg=x and yimg=y in these 2 equations, wee need to have
	* 0 = cos(angle)*tx'-sin(angle)*ty'+tx
	* 0 = sin(angle)*tx'+cos(angle)*ty'+ty
	* This is a 2x2 system of equations for txprime and typrime, so
	* multiply from the left with R' to get
	* cos(angle)*tx'-sin(angle)*ty' = -tx
	* sin(angle)*tx'+cos(angle)*ty' = -ty
	* tx' = cos(angle)*(-tx)+sin(angle)*(-ty)
	* ty' = -sin(angle)*(-tx)+cos(angle)*(-ty)
	* And this is the same expression as for the left inverse.
	* (Summary: M are operations that form a group, and the group inverse
	* is unique)
	*/

	double txprime = sc[1]*trans.coord[0] +sc[0]*trans.coord[1] ;
	double typrime = -sc[0]*trans.coord[0] +sc[1]*trans.coord[1] ;
	return RotTrans2D(-angle, -txprime, -typrime) ;
} /* inv */
