/** @file
* The program fitsImg2Asc converts the image in the primary HDU of a FITS file to an ASCII format.
*
*/

/*
* MPIA - GEIRS project
*
* "@(#) $Id$"
*
*/

#include <limits>
#include <unistd.h>
#include "config.h"
#include "FitsImg2Asc.h"

/** Print a usage syntax message detailing the command line
* @param[in] argv0 The name of the main program.
* @since 2012-10-18
*/
void usage(char *argv0)
{
	cout << "Usage:" << endl ;
	cout << argv0 << " fitsIn.fits > fitsout.plt" << endl ;
	cout << argv0 << " -r [xstrt:xend,ystrt:yend] fitsIn.fits > fitsout.plt" << endl ;
	cout << argv0 << " -h [-l] [-N bins] [-m minvalue] [-M maxvalue] [-a [xmin:xmax,ymin:ymax]] [-t out.txt] [-o out.eps] fitsIn.fits [fitsIn.fits ...]" << endl ;
	cout << argv0 << " -b [-i] [-a [xmin:xmax,ymin:ymax]] -m minvalue -M maxvalue [-X] [-Y] fitsIn.fits fitsOut.fits" << endl ;
	cout << "The argument(s) fitsIn is a single existing FITS file name to be read." << endl ;
	cout << "The option -r with a range argument delimits the pixel range to be exported." << endl ;
	cout << "Redirection is into an ASCII file for use with gnuplot(1)." << endl ;
	cout << "Option -h plots a histogram of pixel ADU values." << endl ;
	cout << "Option -b creates a bad pixel mask with ADU values outside the range." << endl ;
} /* usage */

/** The program fitsImg2Asc converts the image in the primary HDU of a FITS file to an ASCII format.
* The input file is an image in the FITS file, of which only the first slice
* is taken if this is a FITS cube.
*
* - 3D view
*    The standard syntax is
*
*    <code>
*    fitsImg2Asc fitsIn.fits > fitsout.plt
*    </code>
*
*    or
*
*    <code>
*    fitsImg2Asc -r '[xstrt:xend,ystrt:yend]' fitsIn.fits > fitsout.plt
*    </code>
*
*    which means that the command line argument is the name of the FITS file
*    with the image, and the output is redirected into some other file.
*    This output file will be roughly a factor of 4 larger than the input file,
*    and the program will run slowlier than some people would expect.
*    In almost all cases the field will be too crowded to get useful response
*    times from gnuplot while rotating the image, so the option -r allows
*    to take only a rectangular subarea of the image, where the 4 arguments
*    are 0-based ranges for the two pixel axes.
*
*    The output contains lines with three values, which is the x coordinate of
*    the pixel, the y coordinate of the pixel, and the value in the image at
*    this pixel.
*
*    This ASCII file has been targeted for use with a gnuplot(1) session e.g. like:
*    gnuplot
*    gnuplot> set xlabel "x pixel"
*    gnuplot> set ylabel "y pixel"
*    gnuplot> set zlabel "adu"
*    gnuplot> splot "...." wi li  # insert the fitsout.plt file data name here
*    gnuplot> set grid
*    gnuplot> set contour base
*    gnuplot> replot
*    gnuplot> set logscale z
*    gnuplot> replot
*
*    Example:
*    <code>
*    fitsImg2Asc -r '[200:800,1200:1800]' fitsIn.fits > fitsout.plt
*    </code>
*
* - Histogram
*    The syntax to generate a gnuplot X11 window or EPS file with a histogram of
*    pixel ADU values is triggered by the -h option as follows:
*
*    <code>
*    fitsImg2Asc -h [-l] [-N bins] [-m min] [-M max] [-a [xmin:xmax,ymin:ymax]] [-o fitsout.eps] fitsIn.fits [fitsIn.fits ... ]
*    </code>
*
*    The option -N followed by a positive integer number can be used to specify
*    the number of bins to be generated. If not provided, the program uses
*    a default.
*
*    The option -m followed by a number is the minimum number on the horizontal axis
*    which delimits the range of ADU's to be shown.
*
*    The option -M followed by a number is the maximum number on the horizontal axis
*    which delimits the range of ADU's to be shown.
*
*    The option -l means a logarithmic scale is used on the vertical axis.
*
*    The option -a followed by four positive integer numbers in brackets selects
*    a quadrangular region (of FITS coordinates) in the image to be scanned.
*    The default is to scan all pixels in the image.
*
*    The option -o followed by a file name causes the plot to be moved to a EPS file,
*    not to the screen.
*
*    The option -t followed by a file name causes the histogram to be dumped into the named ASCII file.
*
*    The other command line arguments are existing fits files which will be read.
*    The gnuplot display shows the count of pixels on a per-file basis which
*    fall into a range of ADU's. There are two variants of showing the result,
*    one with a logarithmic scale for the counts in the bins, one with a linear
*    scale.
*   
*    Example (which uses the shell's expansion mechanism of file names):   
*   
*    fitsImg2Asc -h -m 0 -M 200 aa000[1-8].fits
*    fitsImg2Asc -h -m '-50' -M 50 -N 100 -a '[5:2044,5:2044]' aa000[1-8].fits
*
*    Example (which shows the selection mechanism of moving to a named
*    extension HDU to locate the image):   
*
*    fitsImg2Asc -h -m '-50' -M 50 -N 100 -a '[5:2044,5:2044]' aa0001.fits'[WIN1]'
*
* - Bad pixel mask
*    The syntax to generate a bad pixel mask is:
*
*    <code>
*    fitsImg2Asc -b [-a '[xstrt:xend,ystrt:yend]'] [-t textout] [-X] [-Y] -m minADU -M maxADU fitsIn.fits fitsOut.fits
*    </code>
*
*    The mask is defined by investigating the 2D image provided in the input
*    file fitsIn.fits and writing the bad pixel mask to fitsOut.fits.
*
*    The option -a defines a subwindow of pixel ranges to be scanned for out-of-range
*    values. The pixels inside the original window but outside that internal
*    window are a frame which is also regarded of containing only bad
*    pixels. So usually for a single Hawaii-2 RG image one would declare
*    the 4-pixel frame to be bad with -a '[5:2044,5:2044]' .
*
*    The option -t followed by the name of a file lets the program generate
*    that file where the bad pixels are listed (one by a line) with their
*    1-based integer x and y coordinates. (This is also the convention of
*    the fixpix routine of the IRAF reduction. It does not produce the generalized format
*    with four parameters equivalent to xstrt, xend, ystrt, yend to define
*    bad pixel blocks.)
*
*    The option -X and similarly the option -Y trigger that the bad pixel
*    mask is generated after flipping the image of fitsIn.fits left-right
*    along x or up-down along y. This helps to generate the mask for
*    images that had another WCS orientation convention than the images
*    the bad pixel mask will be generated for.
*
*    The option -m (quasi mandatory) defines the minimum ADU value and the
*    option -M (quasi mandatory) defines the maximum ADU value of pixels
*    in fitsIn.fits supposed to be good.
*
*    The option -i causes the bad pixel mask to be written with values of 0
*    for good pixels and values of 1 for bad pixels (as in some IRAF conventions).
*    The default (without the option) is to write 1's for good values and
*    NaN for bad values, so it could be used for multiplicative application
*    for generic pictures.
*
*    The two final arguments are the file names of the input image and of the
*    bad pixel image to be created.
*
* @param[in] argc The number of command line switches and arguments.
* @param[in] argv The vector of all command line arguments.
* @since 2012-11-25
* @since 2013-11-11 Support histograms of images in extension headers
* @since 2013-12-04 Implemented bad pixel mask creation
* @see also http://heasarc.gsfc.nasa.gov/ftools/caldb/help/ftstat.html
*/
int main(int argc, char *argv[])
{
	/* option character
	*/
	char oc ;

	string range ;

	string epsout ;
	string txtout ;

	string fitsout ;

	bool doHist = false;
	bool doBad = false;
	bool doLogHist = false;
	bool iraf = false;

	int Nbin = -1 ;

	/* no flips by default
	*/
	int flipxy = 0 ;

	double binRange[2] ;
	
	binRange[0] = binRange[1] = numeric_limits<double>::quiet_NaN() ;

	string detsize = "" ;

	while (  (oc=getopt(argc,argv,"r:hN:m:M:a:lo:t:bXYi")) != -1 )
	{
		switch(oc)
		{
		case 'r' :
			range = string(optarg) ;
			break ;
		case 'o' :
			epsout = string(optarg) ;
			break ;
		case 't' :
			txtout = string(optarg) ;
			break ;
		case 'h' :
			doHist = true ;
			break ;
		case 'b' :
			doBad = true ;
			break ;
		case 'X' :
			flipxy |= 1 ;
			break ;
		case 'Y' :
			flipxy |= 2 ;
			break ;
		case 'l' :
			doLogHist = true ;
			break ;
		case 'i' :
			iraf = true ;
			break ;
		case 'N' :
			Nbin = atoi(optarg) ;
			break ;
		case 'm' :
			binRange[0] = atof(optarg) ;
			break ;
		case 'M' :
			binRange[1] = atof(optarg) ;
			break ;
		case 'a' :
			detsize = string(optarg) ;
			break ;
		case '?' :
			cerr << "Invalid command line option " << optopt << endl ;
			break ;
		}
	}

	/* ensure that this is called with at least one file name
	*/
	if ( optind >= argc )
	{
		usage(argv[0]) ;
		return 1 ;
	}

	if ( doHist)
	{
		/* collect the file names without any further processing.
		*/
		FitsImg2Asc fin(argc-optind, argv+optind) ;
		/* open the files, collect their ADU's, and bin them.
		*/
		fin.histo(Nbin,binRange,detsize,doLogHist,epsout,txtout) ;
	}
	else if ( doBad)
	{
		FitsImg2Asc fin(argv[optind++]) ;
		fitsout = string(argv[optind]) ;
		/* open the files, collect their ADU's, and bin them.
		*/
		fin.badMask(binRange,detsize,fitsout,flipxy,txtout,iraf) ;
	}
	else
	{
		FitsImg2Asc fin(argv[optind]) ;
		if ( range.length() == 0 )
			fin.dump() ;
		else
			fin.dump(range) ;
	}

	return 0 ;
} /* main */
